from octologo.utils import (
    FONTS_DIR,
    color_scheme_names,
    color_schemes,
    font_list,
    get_font_height,
    get_text_size,
    os,
)
from octologo.wizard import SelectQuestion, TextQuestion
from PIL import Image, ImageColor, ImageDraw, ImageFont
from PIL.Image import Image as ImageClass
from PIL.ImageFont import FreeTypeFont
from textual.validation import Number

questions = [
    SelectQuestion(
        "font",
        "Select a font",
        [(font, font) for font in font_list],
        "Iosevka-Nerd-Font-Complete.ttf",
    ),
    SelectQuestion("color", "Select a color scheme", color_scheme_names, "adi1090x"),
    TextQuestion("underline_count", "Lettrs to undrline", [Number(minimum=0)], "1", "1"),
    TextQuestion("padding_x", "Padding x (px)", [Number()], "200", "200"),
    TextQuestion("padding_y", "Padding y (px)", [Number()], "20", "20"),
    TextQuestion("gap", "Gap between text and bar (px)", [Number()], "20", "20"),
    TextQuestion("bar_size", "Bar weight (px)", [Number()], "20", "20"),
    TextQuestion("additionnal_bar_width", "Additionnal bar width (px)", [Number()], "20", "20"),
]

active = False


def get_image(answers: dict) -> ImageClass:
    # Load the selected font
    font_size = 500
    font: FreeTypeFont = ImageFont.truetype(os.path.join(FONTS_DIR, answers["font"]), font_size)

    # Set the colors
    background = ImageColor.getrgb(color_schemes[answers["color"]]["background"])
    text = ImageColor.getrgb(color_schemes[answers["color"]]["text"])
    accent = ImageColor.getrgb(color_schemes[answers["color"]]["accent"])

    # Get the width and height of the texts
    text_width = get_text_size(answers["name"], font)[0]
    font_height = get_font_height(font)

    # Get the correct image width and height
    image_width = 2 * int(answers["padding_x"]) + text_width
    image_height = 2 * int(answers["padding_y"]) + font_height

    # Create the image
    image = Image.new("RGB", (image_width, image_height), background)
    draw = ImageDraw.Draw(image)

    # Get the text anchor type and position on the image (where the text will be drawn)
    # LM = Left/Middle
    anchor_type = "lm"
    anchor_x = int(answers["padding_x"])
    anchor_y = image_height / 2 - (int(answers["gap"]) + int(answers["bar_size"])) / 2

    anchor_pos = (anchor_x, anchor_y)

    if int(answers["underline_count"]) > 0:
        # Get the bbox of the first n letter to underline
        first_letters_bbox = draw.textbbox(
            anchor_pos,
            answers["name"][: int(answers["underline_count"])],
            font=font,
            anchor=anchor_type,
        )

        # Get the underline position
        underline_start_x = first_letters_bbox[0] - int(answers["additionnal_bar_width"])
        underline_start_y = first_letters_bbox[3] + int(answers["gap"])

        underline_end_x = int(answers["additionnal_bar_width"]) + first_letters_bbox[2]

        underline_end_y = underline_start_y + int(answers["bar_size"])

        underline_start = (underline_start_x, underline_start_y)
        underline_end = (underline_end_x, underline_end_y)

        underline_pos = (underline_start, underline_end)

        # Underline the first letter
        draw.rectangle(underline_pos, fill=accent, width=answers["bar_size"])

    # Draw the text
    draw.text(
        anchor_pos,
        answers["name"],
        font=font,
        fill=text,
        anchor=anchor_type,
    )

    # Redraw the first letter
    draw.text(
        anchor_pos,
        answers["name"][0],
        font=font,
        fill=accent,
        anchor=anchor_type,
    )

    return image



from PIL.Image import Image

from . import underline_core

display_name = "First letters underlined"
active = True
questions = underline_core.questions


def get_image(answers: dict) -> Image:
    return underline_core.get_image(answers)



from collections.abc import Generator
from typing import Any

import inquirer as inq
from textual import on
from textual.message import Message
from textual.reactive import reactive
from textual.validation import ValidationResult, Validator
from textual.widgets import Button, Input, Label, Select, Static

from octologo.utils import logger

# from textual import log


class QuestionBase:
    name: str
    label: str

    def __init__(self, name: str, label: str) -> None:
        self.name = name
        self.label = label


class TextQuestion(QuestionBase):
    validators: list[Validator] | None
    placeholder: str
    default_value: str

    def __init__(
        self,
        name: str,
        label: str,
        validators: list[Validator] | None = None,
        placeholder: str = "",
        default_value: str = "",
    ) -> None:
        super().__init__(name, label)
        self.validators = validators
        self.placeholder = placeholder
        self.default_value = default_value

    def as_widget(self) -> Input:
        """Returns a Textual input widget with the correcponding information"""
        _input = Input(
            classes="full-width",
            id=self.name,
            placeholder=self.placeholder,
            validators=self.validators,
        )
        _input.border_title = self.label
        _input.value = self.default_value
        return _input


class SelectQuestion(QuestionBase):
    options: list[tuple[str, Any]]
    default_value: Any | None = None

    def __init__(self, name: str, label: str, options: list[tuple[str, Any]], default_value: str | None = None) -> None:
        super().__init__(name, label)
        self.options = options
        self.default_value = default_value

    def as_widget(self) -> Select:
        _select = Select(
            classes="full-width",
            id=self.name,
            options=self.options,
            allow_blank=False,
            value=self.default_value,
        )

        _select.border_title = self.label

        return _select


class BackNextButtons(Static):
    def compose(self) -> Generator:
        yield Button("Back", variant="warning", id="back")
        yield Button("Next", variant="success", id="next")


class Wizard(Static):
    question_index = reactive(-1)
    answers: dict = dict()
    selected_question: None | Select | Input = None
    questions: list[SelectQuestion | TextQuestion]
    input_message: Label
    title: str = "Wizard"

    # The message sent when the "next" button is clicked while on the last question
    class Finished(Message):
        answers: dict
        wizard_id: str | None

        def __init__(self, answers: dict, wizard_id: str | None) -> None:
            self.answers = answers
            self.wizard_id = wizard_id
            super().__init__()

    @on(Button.Pressed, "#back")
    def on_back(self) -> None:
        # When the back button is pressed, just go to the previous question
        # It cannot be pressed if there arent any questions before
        self.question_index -= 1

    @on(Button.Pressed, "#next")
    async def on_next(self) -> None:
        # If the selected question is an input then fire the submit event so that
        # validation is made and the next question is shown.
        # Else, just go to the next question since a select cannot be invalid
        if isinstance(self.selected_question, Input):
            await self.selected_question.action_submit()
        else:
            self.question_index += 1

    def handle_validation_result(self, validation_result: ValidationResult | None) -> None:
        if self.selected_question is None:
            raise Exception("selected_question should not be None")

        if validation_result is None or validation_result.is_valid:
            # If the validation is OK then hide the error message and set the input color back to normal
            # Also, reenable the next button
            self.query_one("#next").disabled = False
            self.selected_question.remove_class("invalid")
            self.input_message.add_class("hidden")

        else:
            # If the validation comports an error then disable the next button,
            # Show and set the content of the error message and set the input's color to red
            self.query_one("#next").disabled = True
            self.input_message.remove_class("hidden")
            self.input_message.renderable = validation_result.failure_descriptions[0]
            self.selected_question.add_class("invalid")

    def on_input_changed(self, message: Input.Changed) -> None:
        # When an input is changed, save its new value into the self.answers dict
        self.answers[message.input.id] = message.value

        # Show error messages if any
        self.handle_validation_result(message.validation_result)

    def on_input_submitted(self, message: Input.Submitted) -> None:
        # Handle the validation result to show
        # a message if there are any errors
        self.handle_validation_result(message.validation_result)

        # When the input is submitted, if it is valid then go to the next question
        if message.validation_result is None or message.validation_result.is_valid:
            self.question_index += 1

    def on_select_changed(self, message: Select.Changed) -> None:
        # When a select is changed update the value in the self.answers dict
        self.answers[message.select.id] = message.value

    def compose(self) -> Generator:
        # Render directly every input
        # They are all hidden by default
        for i, question in enumerate(self.questions):
            wid = question.as_widget()

            # For every select, the value in the answers dict
            # will not be updated if the user just keeps the default value
            # and click next without changing the value,
            # the on_select_changed function will not be called and the
            # answers dict will not contain the key corresponding to the select which can result in a KeyError
            if isinstance(wid, Select):
                self.answers[wid.id] = question.default_value

            # Hide every questions except the first one
            if i != 0:
                wid.add_class("hidden")

            # mount the question
            yield wid

        # The error message below inputs if there are any errors
        self.input_message = Label("This is the input error message", id="input_message", classes="hidden")
        self.input_message.styles.color = "tomato"
        self.input_message.styles.max_width = "100%"
        yield self.input_message

        # ----------------------------
        yield BackNextButtons()

    def on_mount(self) -> None:
        # Trigger the watch_question_index function to make the first input appear
        self.question_index = 0

    def watch_question_index(self) -> None:
        """Called when the question index changes"""

        # Add 'hidden' class to the previous shown input if any
        if self.selected_question is not None:
            self.selected_question.add_class("hidden")

        # If the question index has been incremented but it is now out of bound then
        # the user clicked next on the last question
        if self.question_index >= len(self.questions):
            self.post_message(self.Finished(self.answers, self.id))
            return

        # Put the question index in the border title
        self.border_title = f"{self.title} [{self.question_index + 1}/{len(self.questions)}]"

        # Show the input corresponding to the new value of self.question_index
        element = self.query_one(f"#{self.questions[self.question_index].name}")
        if not isinstance(element, (Input, Select)):
            raise Exception("Wanted Input or Select but got something else")
        self.selected_question = element
        self.selected_question.remove_class("hidden")
        self.selected_question.focus()

        # If the first question has just been selected then disable
        # the "Back" button since there arent any questions before
        self.query_one("#back").disabled = self.question_index == 0


def validate_text_question(question: TextQuestion, value: str) -> bool:
    if question.validators is None:
        return True

    for validator in question.validators:
        validation = validator.validate(value)
        if not validation.is_valid:
            logger.error(validation.failure_descriptions[0])
            return False

    return True


def inq_ask(questions: list[SelectQuestion | TextQuestion]) -> dict[str, Any]:
    answers = dict()

    for question in questions:
        if isinstance(question, SelectQuestion):
            answers[question.name] = inq.list_input(
                question.label, choices=question.options, default=question.default_value
            )

        elif isinstance(question, TextQuestion):
            while True:
                temp = inq.text(question.label, default=question.default_value)
                if not validate_text_question(question, temp):
                    continue

                answers[question.name] = temp
                break

    return answers



import os
from importlib import import_module
from sys import stdout
from time import time
from types import ModuleType
from typing import Any

import toml
from loguru import logger
from PIL import Image, ImageDraw
from PIL.ImageFont import FreeTypeFont

logger.remove()
logger.add(
    stdout,
    format="[ <green>{time:HH:mm:ss}</green> ]" " - <level>{level}</level> -> " "<level>{message}</level>",
)


def get_text_size(text: str, font: FreeTypeFont) -> tuple[int, int]:
    text_bbox = ImageDraw.Draw(Image.new("RGBA", (1, 1), (0, 0, 0, 0))).textbbox((0, 0), text, font=font)
    text_width = text_bbox[2] - text_bbox[0]
    text_height = text_bbox[3] - text_bbox[1]

    return text_width, text_height


def get_font_height(font: FreeTypeFont) -> int:
    return font.getbbox("azertyuiopqsdfghjklmwxcvbnAZERTYUIOPQASDFGHJKLMWXCVBN0123456789")[3]


def remove_ext(filename: str) -> str:
    """
    Remove the extension from a filename if there is one
    """
    return filename.split(".")[0]


class Style:
    display_name: str
    module: ModuleType

    def __init__(self, display_name: str, module: ModuleType) -> None:
        self.display_name = display_name
        self.module = module


BASE_DIR = os.path.dirname(os.path.abspath(__file__))
FONTS_DIR = os.path.join(BASE_DIR, "fonts")
COLORS_DIR = os.path.join(BASE_DIR, "colors")


# Get all the fonts in the fonts directory
def get_font_list() -> list[str]:
    return os.listdir(FONTS_DIR)


font_list = get_font_list()


# Get all the color schemes in the colors directory
# keep only files with the .toml extension
# and remove the extension
def get_color_schemes() -> dict[str, dict[str, str]]:
    colors = dict()
    for color_file_name in os.listdir(COLORS_DIR):
        # Only keep .toml files
        if not color_file_name.endswith(".toml"):
            continue

        color_parsed = toml.load(os.path.join(COLORS_DIR, color_file_name))

        colors[remove_ext(color_file_name)] = color_parsed

    return colors


color_schemes = get_color_schemes()
color_scheme_names: list[tuple[str, Any]] = [
    (color_schemes[color_scheme]["name"], color_scheme) for color_scheme in color_schemes
]


def get_styles() -> dict[str, Style]:
    """
    return an array of style_id => Style {.module, .display_name}
    """
    # Load the styles in the styles directory
    styles = dict()
    for style in os.listdir(os.path.join(BASE_DIR, "styles")):
        # Only keep .py files
        if not style.endswith(".py"):
            continue

        module = import_module(f"octologo.styles.{remove_ext(style)}")

        # Only keep files with the active attribute set to True
        # This allows to ignore some scripts that may be in the styles directory
        if not module.active:
            continue

        styles[remove_ext(style)] = Style(module.display_name, module)

    return styles


def get_output_filename(project_name: str) -> str:
    return f"octologo_{project_name}_{int(time())}.png"


styles = get_styles()

style_names: list[tuple[str, Any]] = [(styles[style].display_name, style) for style in styles]



# ---------------------------------------- Standard Imports
import os
from collections.abc import Generator

# ---------------------------------------- External Imports
from click_extra import ExtraContext, Parameter, extra_command, option
from PIL.Image import Image
from textual.app import App
from textual.events import Key
from textual.validation import Length
from textual.widgets import Footer, Header, Label, LoadingIndicator

# ---------------------------------------- Local Imports
from octologo import __version__
from octologo.utils import BASE_DIR, get_output_filename, logger, style_names, styles
from octologo.wizard import SelectQuestion, TextQuestion, Wizard, inq_ask

BASIC_INFO_QUESTIONS = [
    TextQuestion(
        "name",
        "Your project's name",
        [Length(1, failure_description="Your project's name cannot be blank")],
        "super-octo-project",
    ),
    SelectQuestion("style", "Logo Style", style_names, "first_letter_underlined"),
]


class OctoLogoApp(App):
    BINDINGS = [
        ("ctrl+q", "quit", "Quit"),
        ("ctrl+t", "toggle_dark", "Toggle Dark Mode"),
    ]
    answers = dict()

    CSS_PATH = os.path.join(BASE_DIR, "app.tcss")
    TITLE = "Octo Logo Wizard"
    finished: bool = False
    result: Image | None = None
    loading_wid: LoadingIndicator = LoadingIndicator(classes="hidden")

    async def on_key(self, event: Key) -> None:
        if event.key == "enter" and self.finished:
            await self.action_quit()
        elif event.key == "v" and self.finished:
            if self.result is None:
                raise Exception("self.result should not be null")
            self.result.show()

    def on_wizard_finished(self, message: Wizard.Finished) -> None:
        # Get the wizard answers and the wizard's id
        self.answers.update(message.answers)
        finished_wizard_id = message.wizard_id

        # remove the wizard
        self.query_one(f"#{finished_wizard_id}").remove()

        # When the basic info wizard is finished, mount the style-specific wizard
        if finished_wizard_id == "basic_info_wizard":
            style_wizard = Wizard(id="style_wizard")
            style_wizard.questions = styles[self.answers["style"]].module.questions
            style_wizard.title = "Style Settings"
            self.mount(style_wizard)
        # When the style-specific wizard is finished, create the image and save it
        elif finished_wizard_id == "style_wizard":
            style = styles[self.answers["style"]].module
            self.result = style.get_image(self.answers)
            self.loading_wid.remove_class("hidden")
            self.set_timer(2, self.final_message)

    # Final message
    def final_message(self) -> None:
        if self.result is None:
            raise Exception("self.result should not be none in OctoLogoApp.final_message")

        save_to = get_output_filename(self.answers["name"])

        self.loading_wid.add_class("hidden")
        self.mount(
            Label(
                f"Logo saved to [bold]{save_to}[/bold].\n"
                f"[blue blink]-> Press v to view the result[/blue blink]\n"
                f"[red]Press enter to quit[/red]"
            )
        )
        self.result.save(save_to)
        self.finished = True

    def compose(self) -> Generator:
        self.app.title = f"Octo Logo v{__version__}"

        yield Header(show_clock=True)
        yield Footer()

        basic_info_wizard = Wizard(id="basic_info_wizard")
        basic_info_wizard.questions = BASIC_INFO_QUESTIONS
        basic_info_wizard.title = "Basic Information"
        yield basic_info_wizard
        yield self.loading_wid


def disable_ansi(ctx: ExtraContext, _: Parameter, val: bool) -> bool:
    ctx.color = not val

    # We must return the value for the main function no_ansi parameter not to be None
    return val


@extra_command(params=[])
@option("-t", "--no-tui", is_flag=True, help="Dont use the Textual Terminal User Interface")
def main(no_tui: bool) -> None:
    if no_tui:
        # If the tui is disabled, do everything without textual
        answers = dict()

        answers.update(inq_ask(BASIC_INFO_QUESTIONS))
        answers.update(inq_ask(styles[answers["style"]].module.questions))

        style = styles[answers["style"]].module
        result = style.get_image(answers)

        save_to = get_output_filename(answers["name"])
        result.save(save_to)
        logger.success(f"Image saved to : {save_to}")
    else:
        # If the tui is enabled, run the textual app
        app = OctoLogoApp()
        app.run()
        quit(0)
